/* hw/s3c2410x_gpio.c
 *
 * Samsung S3C2410X emulation
 *
 * Copyright 2006, 2007 Daniel Silverstone and Vincent Sanders
 *
 * This file is under the terms of the GNU General Public
 * License Version 2
 */

#include "vl.h"
#include "s3c2410x.h"
#include "hw/irq.h"

static uint32_t s3c2410x_gpio_reg[47];

#define EINT_MASK (0xA4)
#define EINT_PEND (0xA8)
#define GPR(P) s3c2410x_gpio_reg[P>>2]

static void
s3c2410x_gpio_propogate_eint(qemu_irq *real_irqs)
{
  uint32_t ints, i;
  S3C2410X_DBF("EINT: Propogating to IRQ\n");
  ints = GPR(EINT_PEND) & ~GPR(EINT_MASK);
  /* EINT0 - EINT3 are INT0 - INT3 */
  for(i=0; i < 4; ++i)
    qemu_set_irq(real_irqs[i], (ints&(1<<i))?1:0);
  /* EINT4 - EINT7 are INT4 */
  qemu_set_irq(real_irqs[4], (ints&0xf0)?1:0);
  /* EINT8 - EINT23 are INT5 */
  qemu_set_irq(real_irqs[5], (ints&0x00ffff00)?1:0);
}

static uint32_t
gpio_con_to_mask(uint32_t con)
{
  uint32_t mask = 0x0;
  int bit;

  for (bit = 0; bit < 16; bit++) {
    if (((con >> (bit*2)) & 0x3) == 0x01)
      mask |= 1 << bit;
  }

  S3C2410X_DBF("Con %08x => mask %08x\n", con, mask);
  return mask;
}

static void
s3c2410x_gpio_write_f(void *opaque, target_phys_addr_t addr_, uint32_t value)
{
  qemu_irq *real_irqs = (qemu_irq *)opaque;
  int addr = (addr_ >> 2) & 0x3f;
  if (addr < 0 || addr > 47) addr = 47;

  if (addr == (EINT_MASK>>2)) value &= ~0xf; /* cannot mask EINT0-EINT3 */
  S3C2410X_DBF("Write GPIO[%02x] = %08x\n",addr<<2, value);
  if (addr == (EINT_PEND>>2))
    s3c2410x_gpio_reg[addr] &= ~value;
  else {
    if (addr < (0x80/4) && (addr_ & 0xf) == 0x04) {
      uint32_t mask = gpio_con_to_mask(s3c2410x_gpio_reg[addr-1]);

      value &= mask;

      s3c2410x_gpio_reg[addr] &= ~mask;
      s3c2410x_gpio_reg[addr] |= value;
    } else
      s3c2410x_gpio_reg[addr] = value;
  }

  if ((addr == (EINT_MASK)>>2) || (addr == (EINT_PEND)>>2)) {
    /* A write to the EINT regs leads us to determine the interrupts to propagate */
    s3c2410x_gpio_propogate_eint(real_irqs);
  }
}

static uint32_t
s3c2410x_gpio_read_f(void *opaque_ignored, target_phys_addr_t addr_)
{
  int addr = (addr_ >> 2) & 0x3f;
  if (addr_ == 0x560000b0) return CPU_S3C2410X_IDENT_S3C2410A;
  if (addr < 0 || addr > 47) addr = 47;
  /* If IIC pins are wired to GPE, and GPE is on input mode, pretend the IIC is pulled high */
  if (addr == 0x44>>2) {
    /* GPEDAT read. */
    uint32_t ret = s3c2410x_gpio_reg[addr];
    if ((s3c2410x_gpio_reg[15] & 3<<28) == 0) ret |= 1 << 14;
    if ((s3c2410x_gpio_reg[15] & 3<<30) == 0) ret |= 1 << 15;
    S3C2410X_DBF("GPIO read of GPE, returning 0x%08x\n", ret);
    return ret;
  }
  S3C2410X_DBF("Read GPIO[%02x] => 0x%08x\n",addr<<2, s3c2410x_gpio_reg[addr]);
  return s3c2410x_gpio_reg[addr];
}


static CPUReadMemoryFunc *s3c2410x_gpio_read[] = {
  &s3c2410x_gpio_read_f,
  &s3c2410x_gpio_read_f,
  &s3c2410x_gpio_read_f,
};

static CPUWriteMemoryFunc *s3c2410x_gpio_write[] = {
  &s3c2410x_gpio_write_f,
  &s3c2410x_gpio_write_f,
  &s3c2410x_gpio_write_f,
};

static void
s3c2410x_gpio_irq_handler(void *opaque, int n, int level)
{
  qemu_irq *real_irqs = (qemu_irq *)opaque;
  if (level)
    GPR(EINT_PEND) |= (1<<n);
  s3c2410x_gpio_propogate_eint(real_irqs);
}

qemu_irq *
s3c2410x_gpio_init(qemu_irq *real_irqs)
{
  /* Samsung S3C2410X GPIO magic badger.
   *
   * The only really magic thing we do currently is the ID register
   */
  int tag = cpu_register_io_memory(0, s3c2410x_gpio_read, s3c2410x_gpio_write, real_irqs);
  cpu_register_physical_memory(CPU_S3C2410X_GPIO_BASE, 47*4, tag);
  GPR(0x00) = 0x7fffff;
  GPR(0x10) = 0;
  GPR(0x18) = 0;
  GPR(0x20) = 0;
  GPR(0x28) = 0;
  GPR(0x30) = 0;
  GPR(0x34) = 0xfefc;		/* This happens to be the default in the h1940 */
  GPR(0x38) = 0xf000;
  GPR(0x40) = 0;
  GPR(0x48) = 0;
  GPR(0x50) = 0;
  GPR(0x58) = 0;
  GPR(0x60) = 0;
  GPR(0x68) = 0xf800;
  GPR(0x70) = 0;
  GPR(0x78) = 0;
  GPR(0x80) = 0x10330;
  GPR(0x84) = 0;
  GPR(0x88) = 0;
  GPR(0x8C) = 0;
  GPR(0x90) = 0;
  GPR(0x9C) = 0;
  GPR(0xA0) = 0;
  GPR(0xA4) = 0xfffff0;
  GPR(0xA8) = 0;
  GPR(0xB0) = 0x3241000;
  GPR(0xB4) = 1;
  GPR(0xB8) = 0;
  GPR(0xBC) = 0;
  /* EINTs 0-23 */
  return qemu_allocate_irqs(s3c2410x_gpio_irq_handler, real_irqs, 24);
}
