/* hw/s3c2410x_iic.c
 *
 * Samsung S3C2410X emulation
 *
 * Copyright 2006, 2007 Daniel Silverstone, Ben Dooks
 *  and Vincent Sanders
 *
 * This file is under the terms of the GNU General Public
 * License Version 2
 */

//#define DEBUG_S3C2410X

#include "vl.h"
#include "s3c2410x.h"
#include "s3c24xx/regs-iic.h"

static uint32_t s3c2410x_iic_reg[4];

#undef S3C2410_IICREG
#define S3C2410_IICREG(x)	(x)

static void
s3c2410x_iic_write_f(void *opaque, target_phys_addr_t addr_, uint32_t value)
{
  qemu_irq iic_irq = (qemu_irq)opaque;
  int addr = (addr_ & 0xf);
  
  S3C2410X_DBF("Write IIC[%02x] = %08x\n", addr, value);


  switch (addr) {
  case S3C2410_IICCON:
    value |= (s3c2410x_iic_reg[0] & S3C2410_IICCON_IRQPEND);
    s3c2410x_iic_reg[0] = value ;
    break;

  case S3C2410_IICSTAT:
    if ((value & S3C2410_IICSTAT_START) &&
	(((value & S3C2410_IICSTAT_MODEMASK) == S3C2410_IICSTAT_MASTER_TX) ||
	 ((value & S3C2410_IICSTAT_MODEMASK) == S3C2410_IICSTAT_MASTER_RX))) {
      /* When we're Master TX and told to start, we hiccough inappropriately */
      S3C2410X_DBF("Oooh, IIC write to fail, yum!\n");

      value &= ~S3C2410_IICSTAT_BUSBUSY; /* Not running */
      s3c2410x_iic_reg[1] = value;

      s3c2410x_iic_reg[0] |= S3C2410_IICCON_IRQPEND; /* Claim an interrupt */

      switch (s3c2410x_iic_reg[1] & S3C2410_IICSTAT_MODEMASK) {
      case S3C2410_IICSTAT_MASTER_TX:
	break;

      case S3C2410_IICSTAT_MASTER_RX:
	break;
      }
      
      S3C2410X_DBF("And raise an interrupt\n");
      qemu_set_irq(iic_irq, 1);
    }
    break;

  case S3C2410_IICDS:
    /* write to data register */
    s3c2410x_iic_reg[addr >> 2] = 0xff;
    break;
    
  default:
    s3c2410x_iic_reg[addr >> 2] = value;
  }
}

static const char *
decode_iicstat_mode(uint32_t mode)
{
  switch (mode & S3C2410_IICSTAT_MODEMASK) {
  case S3C2410_IICSTAT_MASTER_RX:
    return "MasterRX";
  case S3C2410_IICSTAT_MASTER_TX:
    return "MasterTX";
  case S3C2410_IICSTAT_SLAVE_RX:
    return "SlaveRX";
  case S3C2410_IICSTAT_SLAVE_TX:
    return "SlaveTX";
  }

  return "";
}

static uint32_t
s3c2410x_iic_read_f(void *opaque_ignored, target_phys_addr_t addr_)
{
  int addr = (addr_ & 0xf);
  uint32_t val;

  if (addr < 0 || addr > 12) addr = 12;

  val = s3c2410x_iic_reg[addr >> 2];

  S3C2410X_DBF("Read IIC[%02x] => %02x\n", addr, val);

  switch (addr) {
  case S3C2410_IICCON:
    S3C2410X_DBF("Read: IICCON: AckEn %d, TxDiv %d, IrqEn %d, IrqPend %d, Scale %d\n",
		 (val & S3C2410_IICCON_ACKEN) ? 1 : 0,
		 (val & S3C2410_IICCON_TXDIV_512) ? 512 : 16,
		 (val & S3C2410_IICCON_IRQEN) ? 1 : 0,
		 (val & S3C2410_IICCON_IRQPEND) ? 1 : 0,
		 (val & S3C2410_IICCON_SCALEMASK));
    break;

  case S3C2410_IICSTAT:
    S3C2410X_DBF("Read: IICSTAT: Mode %s, BusBusy %d, TxRxEn %d, Arb %d, Slave %d, Addr0 %d, LBit %d\n",
		 decode_iicstat_mode(val),
		 (val & S3C2410_IICSTAT_BUSBUSY) ? 1 : 0,
		 (val & S3C2410_IICSTAT_TXRXEN) ? 1 : 0,
		 (val & S3C2410_IICSTAT_ARBITR) ? 1 : 0,
		 (val & S3C2410_IICSTAT_ASSLAVE) ? 1 : 0,
		 (val & S3C2410_IICSTAT_ADDR0) ? 1 : 0,
		 (val & S3C2410_IICSTAT_LASTBIT));
  }
  
  return val;
}


static CPUReadMemoryFunc *s3c2410x_iic_read[] = {
  &s3c2410x_iic_read_f,
  &s3c2410x_iic_read_f,
  &s3c2410x_iic_read_f,
};

static CPUWriteMemoryFunc *s3c2410x_iic_write[] = {
  &s3c2410x_iic_write_f,
  &s3c2410x_iic_write_f,
  &s3c2410x_iic_write_f,
};


void
s3c2410x_iic_init(qemu_irq iic_irq)
{
  /* Samsung S3C2410X IIC registration. */
  int tag = cpu_register_io_memory(0, s3c2410x_iic_read, s3c2410x_iic_write, iic_irq);
  cpu_register_physical_memory(CPU_S3C2410X_IIC_BASE, 4*4, tag);
  s3c2410x_iic_reg[0] = 0;
  s3c2410x_iic_reg[1] = 0;
  s3c2410x_iic_reg[2] = 0;
  s3c2410x_iic_reg[3] = 0;
  
}
