/* hw/simtecbast.c
 *
 * Simple system emulation for the Simtec Electronics BAST
 *
 * Copyright 2006 Daniel Silverstone and Vincent Sanders
 *
 * This file is under the terms of the GNU General Public
 * License Version 2.
 */

#include "vl.h"
#include "s3c2410x.h"
#include "assert.h"

#define BIOS_FILENAME "s3c2410x-able.bin"

/* Bytes in a Kilobyte */
#define KILO 1024
/* Bytes in a megabyte */
#define MEGA 1024 * KILO
/* Bytes */
#define BYTE 1
/* Bits in a byte */
#define BIT 8

/* Useful defines */
#define SIMTEC_BAST_NOR_BASE CPU_S3C2410X_CS0
#define SIMTEC_BAST_NOR_SIZE 16 * MEGA / BIT
#define SIMTEC_BAST_BOARD_ID 331

#define BAST_IDE_PRI_SLOW    (CPU_S3C2410X_CS3 | 0x2000000)
#define BAST_IDE_SEC_SLOW    (CPU_S3C2410X_CS3 | 0x3000000)
#define BAST_IDE_PRI_FAST    (CPU_S3C2410X_CS5 | 0x2000000)
#define BAST_IDE_SEC_FAST    (CPU_S3C2410X_CS5 | 0x3000000)

#define BAST_IDE_PRI_SLOW_BYTE    (CPU_S3C2410X_CS2 | 0x2000000)
#define BAST_IDE_SEC_SLOW_BYTE    (CPU_S3C2410X_CS2 | 0x3000000)
#define BAST_IDE_PRI_FAST_BYTE    (CPU_S3C2410X_CS4 | 0x2000000)
#define BAST_IDE_SEC_FAST_BYTE    (CPU_S3C2410X_CS4 | 0x3000000)

static void stcb_register_ide(void)
{
  int ide0_mem = stcb_ide_init(&bs_table[0], s3c_irqs[16]);
  int ide1_mem = stcb_ide_init(&bs_table[2], s3c_irqs[17]);
  
  cpu_register_physical_memory(BAST_IDE_PRI_SLOW, 0x1000000, ide0_mem);
  cpu_register_physical_memory(BAST_IDE_PRI_FAST, 0x1000000, ide0_mem);
  
  cpu_register_physical_memory(BAST_IDE_SEC_SLOW, 0x1000000, ide1_mem);
  cpu_register_physical_memory(BAST_IDE_SEC_FAST, 0x1000000, ide1_mem);

  cpu_register_physical_memory(BAST_IDE_PRI_SLOW_BYTE, 0x1000000, ide0_mem);
  cpu_register_physical_memory(BAST_IDE_PRI_FAST_BYTE, 0x1000000, ide0_mem);
  
  cpu_register_physical_memory(BAST_IDE_SEC_SLOW_BYTE, 0x1000000, ide1_mem);
  cpu_register_physical_memory(BAST_IDE_SEC_FAST_BYTE, 0x1000000, ide1_mem);
}

static void stcb_init(int _ram_size, int vga_ram_size, int boot_device,
                      DisplayState *ds, const char **fd_filename, int snapshot,
                      const char *kernel_filename, const char *kernel_cmdline,
                      const char *initrd_filename)
{
  CPUState *env;
  int bios_offset;
  char buf[1024]; /* URGH */
  int ret;
  NICInfo* nd;

  if (_ram_size > (256 * MEGA * BYTE)) _ram_size = 256 * MEGA * BYTE;
  ram_size = _ram_size;

  bios_offset = ram_size + vga_ram_size;

  env = s3c2410x_init(ram_size);
  
  if( kernel_filename == NULL ) {
    snprintf(buf, sizeof(buf), "%s/%s", bios_dir, BIOS_FILENAME);
    ret = load_image(buf, phys_ram_base + bios_offset);
    if( ret <= 0 ) {
      fprintf(stderr, "qemu: warning, could not load BAST BIOS from %s\n", buf);
    } else {
      fprintf(stdout, "qemu: info, loaded BAST BIOS %d bytes from %s\n", ret, buf);
    }
  } else {
    arm_load_kernel(env, ram_size, kernel_filename, kernel_cmdline,
		    initrd_filename, SIMTEC_BAST_BOARD_ID, CPU_S3C2410X_RAM);
    /* Construct the smallest bootloader(ish) in the world */
    stl_raw(phys_ram_base + bios_offset, 0xE59FF000); /* ldr pc, [pc, #0] (loads from +8) */
    stl_raw(phys_ram_base + bios_offset + 4, 0x00000000);
    stl_raw(phys_ram_base + bios_offset + 8, CPU_S3C2410X_RAM);
    stl_raw(phys_ram_base + bios_offset + 12, 0xE3A0F000);
  }
  
  /* Register the NOR flash ROM */
  cpu_register_physical_memory(SIMTEC_BAST_NOR_BASE, SIMTEC_BAST_NOR_SIZE,
                               bios_offset | IO_MEM_ROM);
  stcb_register_ide();

  nd = &nd_table[0];
  if (!nd->model)
      nd->model = "dm9000";
  if (strcmp(nd->model, "dm9000") == 0) {
      dm9000_init(nd, 0x2D000000, 0x00, 0x40, s3c_irqs[10]);
  }

  

  /* And we're good to go */
}


QEMUMachine simtecbast_machine = {
  "simtecbast",
  "Simtec Electronics BAST (S3C2410X, ARM920T)",
  stcb_init,
};
